@@ -47,6 +47,11 @@
 #define PACKAGE_STRING "patchelf"
 #endif
 
+// This is needed for Windows/mingw
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 static bool debugMode = false;
 
 static bool forceRPath = false;
@@ -81,6 +86,20 @@ static bool hasAllowedPrefix(const std::string & s, const std::vector<std::strin
     return std::any_of(allowedPrefixes.begin(), allowedPrefixes.end(), [&](const std::string & i) { return !s.compare(0, i.size(), i); });
 }
 
+static std::string trim(std::string s)
+{
+    s.erase(s.begin(), std::find_if(s.begin(), s.end(), [](unsigned char ch) { return !std::isspace(ch); }));
+    s.erase(std::find_if(s.rbegin(), s.rend(), [](unsigned char ch) { return !std::isspace(ch); }).base(), s.end());
+
+    return s;
+}
+
+static std::string downcase(std::string s)
+{
+    std::transform(s.begin(), s.end(), s.begin(), [](unsigned char c){ return std::tolower(c); });
+    return s;
+}
+
 /* !!! G++ creates broken code if this function is inlined, don't know
    why... */
 template<ElfFileParams>
@@ -164,7 +183,7 @@ static FileContents readFile(const std::string & fileName,
 
     FileContents contents = std::make_shared<std::vector<unsigned char>>(size);
 
-    int fd = open(fileName.c_str(), O_RDONLY);
+    int fd = open(fileName.c_str(), O_RDONLY | O_BINARY);
     if (fd == -1) throw SysError(fmt("opening '", fileName, "'"));
 
     size_t bytesRead = 0;
@@ -299,7 +318,7 @@ unsigned int ElfFile<ElfFileParamNames>::getPageSize() const
     // requirements. There is no authoritative list of these values. The
     // current list is extracted from GNU gold's source code (abi_pagesize).
     switch (rdi(hdr()->e_machine)) {
-      case EM_SPARC:
+      case EM_IA_64:
       case EM_MIPS:
       case EM_PPC:
       case EM_PPC64:
@@ -307,6 +326,10 @@ unsigned int ElfFile<ElfFileParamNames>::getPageSize() const
       case EM_TILEGX:
       case EM_LOONGARCH:
         return 0x10000;
+      case EM_SPARC: // This should be sparc 32-bit. According to the linux
+                     // kernel 4KB should be also fine, but it seems that solaris is doing 8KB
+      case EM_SPARCV9: /* SPARC64 support */
+        return 0x2000;
       default:
         return 0x1000;
     }
@@ -375,7 +398,7 @@ static void writeFile(const std::string & fileName, const FileContents & content
 {
     debug("writing %s\n", fileName.c_str());
 
-    int fd = open(fileName.c_str(), O_CREAT | O_TRUNC | O_WRONLY, 0777);
+    int fd = open(fileName.c_str(), O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, 0777);
     if (fd == -1)
         error("open");
 
@@ -413,44 +436,83 @@ static uint64_t roundUp(uint64_t n, uint64_t m)
 
 
 template<ElfFileParams>
-void ElfFile<ElfFileParamNames>::shiftFile(unsigned int extraPages, Elf_Addr startPage)
+void ElfFile<ElfFileParamNames>::shiftFile(unsigned int extraPages, size_t startOffset)
 {
-    /* Move the entire contents of the file 'extraPages' pages
-       further. */
+    assert(startOffset >= sizeof(Elf_Ehdr));
+
     unsigned int oldSize = fileContents->size();
+    assert(oldSize > startOffset);
+
+    /* Move the entire contents of the file after 'startOffset' by 'extraPages' pages further. */
     unsigned int shift = extraPages * getPageSize();
     fileContents->resize(oldSize + shift, 0);
-    memmove(fileContents->data() + shift, fileContents->data(), oldSize);
-    memset(fileContents->data() + sizeof(Elf_Ehdr), 0, shift - sizeof(Elf_Ehdr));
+    memmove(fileContents->data() + startOffset + shift, fileContents->data() + startOffset, oldSize - startOffset);
+    memset(fileContents->data() + startOffset, 0, shift);
 
     /* Adjust the ELF header. */
     wri(hdr()->e_phoff, sizeof(Elf_Ehdr));
-    wri(hdr()->e_shoff, rdi(hdr()->e_shoff) + shift);
+    if (rdi(hdr()->e_shoff) >= startOffset)
+        wri(hdr()->e_shoff, rdi(hdr()->e_shoff) + shift);
 
     /* Update the offsets in the section headers. */
-    for (int i = 1; i < rdi(hdr()->e_shnum); ++i)
-        wri(shdrs.at(i).sh_offset, rdi(shdrs.at(i).sh_offset) + shift);
+    for (int i = 1; i < rdi(hdr()->e_shnum); ++i) {
+        size_t sh_offset = rdi(shdrs.at(i).sh_offset);
+        if (sh_offset >= startOffset)
+            wri(shdrs.at(i).sh_offset, sh_offset + shift);
+    }
+
+    int splitIndex = -1;
+    size_t splitShift = 0;
 
     /* Update the offsets in the program headers. */
     for (int i = 0; i < rdi(hdr()->e_phnum); ++i) {
-        wri(phdrs.at(i).p_offset, rdi(phdrs.at(i).p_offset) + shift);
-        if (rdi(phdrs.at(i).p_align) != 0 &&
-            (rdi(phdrs.at(i).p_vaddr) - rdi(phdrs.at(i).p_offset)) % rdi(phdrs.at(i).p_align) != 0) {
-            debug("changing alignment of program header %d from %d to %d\n", i,
-                rdi(phdrs.at(i).p_align), getPageSize());
-            wri(phdrs.at(i).p_align, getPageSize());
+        Elf_Off p_start = rdi(phdrs.at(i).p_offset);
+
+        if (p_start <= startOffset && p_start + rdi(phdrs.at(i).p_filesz) > startOffset && rdi(phdrs.at(i).p_type) == PT_LOAD) {
+            assert(splitIndex == -1);
+
+            splitIndex = i;
+            splitShift = startOffset - p_start;
+
+            /* This is the load segment we're currently extending within, so we split it. */
+            wri(phdrs.at(i).p_offset, startOffset);
+            wri(phdrs.at(i).p_memsz, rdi(phdrs.at(i).p_memsz) - splitShift);
+            wri(phdrs.at(i).p_filesz, rdi(phdrs.at(i).p_filesz) - splitShift);
+            wri(phdrs.at(i).p_paddr, rdi(phdrs.at(i).p_paddr) + splitShift);
+            wri(phdrs.at(i).p_vaddr, rdi(phdrs.at(i).p_vaddr) + splitShift);
+
+            p_start = startOffset;
+        }
+
+        if (p_start >= startOffset) {
+            wri(phdrs.at(i).p_offset, p_start + shift);
+            if (rdi(phdrs.at(i).p_align) != 0 &&
+                (rdi(phdrs.at(i).p_vaddr) - rdi(phdrs.at(i).p_offset)) % rdi(phdrs.at(i).p_align) != 0) {
+                debug("changing alignment of program header %d from %d to %d\n", i,
+                    rdi(phdrs.at(i).p_align), getPageSize());
+                wri(phdrs.at(i).p_align, getPageSize());
+            }
+        } else {
+            /* If we're not physically shifting, shift back virtual memory. */
+            if (rdi(phdrs.at(i).p_paddr) >= shift)
+                wri(phdrs.at(i).p_paddr, rdi(phdrs.at(i).p_paddr) - shift);
+            if (rdi(phdrs.at(i).p_vaddr) >= shift)
+                wri(phdrs.at(i).p_vaddr, rdi(phdrs.at(i).p_vaddr) - shift);
         }
     }
 
+    assert(splitIndex != -1);
+
     /* Add a segment that maps the new program/section headers and
        PT_INTERP segment into memory.  Otherwise glibc will choke. */
     phdrs.resize(rdi(hdr()->e_phnum) + 1);
     wri(hdr()->e_phnum, rdi(hdr()->e_phnum) + 1);
     Elf_Phdr & phdr = phdrs.at(rdi(hdr()->e_phnum) - 1);
     wri(phdr.p_type, PT_LOAD);
-    wri(phdr.p_offset, 0);
-    wri(phdr.p_vaddr, wri(phdr.p_paddr, startPage));
-    wri(phdr.p_filesz, wri(phdr.p_memsz, shift));
+    wri(phdr.p_offset, phdrs.at(splitIndex).p_offset - splitShift - shift);
+    wri(phdr.p_paddr, phdrs.at(splitIndex).p_paddr - splitShift - shift);
+    wri(phdr.p_vaddr, phdrs.at(splitIndex).p_vaddr - splitShift - shift);
+    wri(phdr.p_filesz, wri(phdr.p_memsz, splitShift + shift));
     wri(phdr.p_flags, PF_R | PF_W);
     wri(phdr.p_align, getPageSize());
 }
@@ -542,20 +604,26 @@ void ElfFile<ElfFileParamNames>::writeReplacedSections(Elf_Off & curOff,
     }
 
     std::set<unsigned int> noted_phdrs = {};
-    for (auto & i : replacedSections) {
-        const std::string & sectionName = i.first;
-        auto & shdr = findSectionHeader(sectionName);
+
+    /* We iterate over the sorted section headers here, so that the relative
+       position between replaced sections stays the same.  */
+    for (auto & shdr : shdrs) {
+        std::string sectionName = getSectionName(shdr);
+        auto i = replacedSections.find(sectionName);
+        if (i == replacedSections.end())
+            continue;
+
         Elf_Shdr orig_shdr = shdr;
         debug("rewriting section '%s' from offset 0x%x (size %d) to offset 0x%x (size %d)\n",
-            sectionName.c_str(), rdi(shdr.sh_offset), rdi(shdr.sh_size), curOff, i.second.size());
+            sectionName.c_str(), rdi(shdr.sh_offset), rdi(shdr.sh_size), curOff, i->second.size());
 
-        memcpy(fileContents->data() + curOff, (unsigned char *) i.second.c_str(),
-            i.second.size());
+        memcpy(fileContents->data() + curOff, (unsigned char *) i->second.c_str(),
+            i->second.size());
 
         /* Update the section header for this section. */
         wri(shdr.sh_offset, curOff);
         wri(shdr.sh_addr, startAddr + (curOff - startOffset));
-        wri(shdr.sh_size, i.second.size());
+        wri(shdr.sh_size, i->second.size());
         wri(shdr.sh_addralign, sectionAlignment);
 
         /* If this is the .interp section, then the PT_INTERP segment
@@ -633,7 +701,19 @@ void ElfFile<ElfFileParamNames>::writeReplacedSections(Elf_Off & curOff,
             }
         }
 
-        curOff += roundUp(i.second.size(), sectionAlignment);
+        /* If there is .note.gnu.property section, then the PT_GNU_PROPERTY
+           segment must be sync'ed with it. */
+        if (sectionName == ".note.gnu.property") {
+            for (auto & phdr : phdrs) {
+                if (rdi(phdr.p_type) == PT_GNU_PROPERTY) {
+                    phdr.p_offset = shdr.sh_offset;
+                    phdr.p_vaddr = phdr.p_paddr = shdr.sh_addr;
+                    phdr.p_filesz = phdr.p_memsz = shdr.sh_size;
+                }
+            }
+        }
+
+        curOff += roundUp(i->second.size(), sectionAlignment);
     }
 
     replacedSections.clear();
@@ -670,7 +750,7 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsLibrary()
     /* Some sections may already be replaced so account for that */
     unsigned int i = 1;
     Elf_Addr pht_size = sizeof(Elf_Ehdr) + (phdrs.size() + num_notes + 1)*sizeof(Elf_Phdr);
-    while( rdi(shdrs.at(i).sh_offset) <= pht_size && i < rdi(hdr()->e_shnum) ) {
+    while( i < rdi(hdr()->e_shnum) && rdi(shdrs.at(i).sh_offset) <= pht_size ) {
         if (not haveReplacedSection(getSectionName(shdrs.at(i))))
             replaceSection(getSectionName(shdrs.at(i)), rdi(shdrs.at(i).sh_size));
         i++;
@@ -684,7 +764,11 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsLibrary()
 
     Elf_Off startOffset = roundUp(fileContents->size(), getPageSize());
 
-    fileContents->resize(startOffset + neededSpace, 0);
+    // In older version of binutils (2.30), readelf would check if the dynamic
+    // section segment is strictly smaller than the file (and not same size).
+    // By making it one byte larger, we don't break readelf.
+    off_t binutilsQuirkPadding = 1;
+    fileContents->resize(startOffset + neededSpace + binutilsQuirkPadding, 0);
 
     /* Even though this file is of type ET_DYN, it could actually be
        an executable.  For instance, Gold produces executables marked
@@ -730,14 +814,16 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsLibrary()
     rewriteHeaders(firstPage + rdi(hdr()->e_phoff));
 }
 
+static bool noSort = false;
 
 template<ElfFileParams>
 void ElfFile<ElfFileParamNames>::rewriteSectionsExecutable()
 {
-    /* Sort the sections by offset, otherwise we won't correctly find
-       all the sections before the last replaced section. */
-    sortShdrs();
-
+    if (!noSort) {
+        /* Sort the sections by offset, otherwise we won't correctly find
+           all the sections before the last replaced section. */
+        sortShdrs();
+    }
 
     /* What is the index of the last replaced section? */
     unsigned int lastReplaced = 0;
@@ -822,7 +908,6 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsExecutable()
        file by the minimum number of pages and adjust internal
        offsets. */
     if (neededSpace > startOffset) {
-
         /* We also need an additional program header, so adjust for that. */
         neededSpace += sizeof(Elf_Phdr);
         debug("needed space is %d\n", neededSpace);
@@ -832,10 +917,10 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsExecutable()
         if (neededPages * getPageSize() > firstPage)
             error("virtual address space underrun!");
 
+        shiftFile(neededPages, startOffset);
+
         firstPage -= neededPages * getPageSize();
         startOffset += neededPages * getPageSize();
-
-        shiftFile(neededPages, firstPage);
     }
 
 
@@ -873,6 +958,14 @@ void ElfFile<ElfFileParamNames>::normalizeNoteSegments()
         size_t start_off = rdi(phdr.p_offset);
         size_t curr_off = start_off;
         size_t end_off = start_off + rdi(phdr.p_filesz);
+
+        /* Binaries produced by older patchelf versions may contain empty PT_NOTE segments.
+           For backwards compatibility, if we find one we should ignore it. */
+        bool empty = std::none_of(shdrs.begin(), shdrs.end(),
+            [&](auto & shdr) { return rdi(shdr.sh_offset) >= start_off && rdi(shdr.sh_offset) < end_off; });
+        if (empty)
+            continue;
+
         while (curr_off < end_off) {
             /* Find a section that starts at the current offset. If we can't
                find one, it means the SHT_NOTE sections weren't contiguous
@@ -950,7 +1043,9 @@ void ElfFile<ElfFileParamNames>::rewriteHeaders(Elf_Addr phdrAddress)
         }
     }
 
-    sortPhdrs();
+    if (!noSort) {
+        sortPhdrs();
+    }
 
     for (unsigned int i = 0; i < phdrs.size(); ++i)
         * ((Elf_Phdr *) (fileContents->data() + rdi(hdr()->e_phoff)) + i) = phdrs.at(i);
@@ -959,7 +1054,9 @@ void ElfFile<ElfFileParamNames>::rewriteHeaders(Elf_Addr phdrAddress)
     /* Rewrite the section header table.  For neatness, keep the
        sections sorted. */
     assert(rdi(hdr()->e_shnum) == shdrs.size());
-    sortShdrs();
+    if (!noSort) {
+        sortShdrs();
+    }
     for (unsigned int i = 1; i < rdi(hdr()->e_shnum); ++i)
         * ((Elf_Shdr *) (fileContents->data() + rdi(hdr()->e_shoff)) + i) = shdrs.at(i);
 
@@ -985,6 +1082,10 @@ void ElfFile<ElfFileParamNames>::rewriteHeaders(Elf_Addr phdrAddress)
                 // some binaries might this section stripped
                 // in which case we just ignore the value.
                 if (shdr) dyn->d_un.d_ptr = (*shdr).get().sh_addr;
+            } else if (d_tag == DT_MIPS_XHASH) {
+                // the .MIPS.xhash section was added to the glibc-ABI
+                // in commit 23c1c256ae7b0f010d0fcaff60682b620887b164
+                dyn->d_un.d_ptr = findSectionHeader(".MIPS.xhash").sh_addr;
             } else if (d_tag == DT_JMPREL) {
                 auto shdr = tryFindSectionHeader(".rel.plt");
                 if (!shdr) shdr = tryFindSectionHeader(".rela.plt");
@@ -1019,6 +1120,16 @@ void ElfFile<ElfFileParamNames>::rewriteHeaders(Elf_Addr phdrAddress)
                    pointer, relative to the address of the tag */
                 auto shdr = tryFindSectionHeader(".rld_map");
                 if (shdr) {
+                    /*
+                     * "When correct, (DT_MIPS_RLD_MAP_REL + tag offset + executable base address) equals DT_MIPS_RLD_MAP"
+                     * -- https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=820334#5
+                     *
+                     * Equivalently,
+                     *
+                     *   DT_MIPS_RLD_MAP_REL + tag offset + executable base address == DT_MIPS_RLD_MAP
+                     *   DT_MIPS_RLD_MAP_REL              + executable base address == DT_MIPS_RLD_MAP - tag_offset
+                     *   DT_MIPS_RLD_MAP_REL                                        == DT_MIPS_RLD_MAP - tag_offset - executable base address
+                     */
                     auto rld_map_addr = findSectionHeader(".rld_map").sh_addr;
                     auto dyn_offset = ((char*)dyn) - ((char*)dyn_table);
                     dyn->d_un.d_ptr = rld_map_addr - dyn_offset - (*shdrDynamic).get().sh_addr;
@@ -1079,6 +1190,68 @@ std::string ElfFile<ElfFileParamNames>::getInterpreter()
     return std::string((char *) fileContents->data() + rdi(shdr.sh_offset), rdi(shdr.sh_size) - 1);
 }
 
+template<ElfFileParams>
+void ElfFile<ElfFileParamNames>::modifyOsAbi(osAbiMode op, const std::string & newOsAbi)
+{
+    unsigned char abi = hdr()->e_ident[EI_OSABI];
+
+    if (op == printOsAbi) {
+        switch (abi) {
+            case 0:  printf("System V\n"); break;
+            case 1:  printf("HP-UX\n"); break;
+            case 2:  printf("NetBSD\n"); break;
+            case 3:  printf("Linux\n"); break;
+            case 4:  printf("GNU Hurd\n"); break;
+            case 6:  printf("Solaris\n"); break;
+            case 7:  printf("AIX\n"); break;
+            case 8:  printf("IRIX\n"); break;
+            case 9:  printf("FreeBSD\n"); break;
+            case 10: printf("Tru64\n"); break;
+            case 12: printf("OpenBSD\n"); break;
+            case 13: printf("OpenVMS\n"); break;
+            default: printf("0x%02X\n", (unsigned int) abi);
+        }
+        return;
+    }
+
+    unsigned char newAbi;
+    std::string nabi = downcase(trim(newOsAbi));
+    if (nabi == "system v" || nabi == "system-v" || nabi == "sysv")
+        newAbi = 0;
+    else if (nabi == "hp-ux")
+        newAbi = 1;
+    else if (nabi == "netbsd")
+        newAbi = 2;
+    else if (nabi == "linux" || nabi == "gnu")
+        newAbi = 3;
+    else if (nabi == "gnu hurd" || nabi == "gnu-hurd" || nabi == "hurd")
+        newAbi = 4;
+    else if (nabi == "solaris")
+        newAbi = 6;
+    else if (nabi == "aix")
+        newAbi = 7;
+    else if (nabi == "irix")
+        newAbi = 8;
+    else if (nabi == "freebsd")
+        newAbi = 9;
+    else if (nabi == "tru64")
+        newAbi = 10;
+    else if (nabi == "openbsd")
+        newAbi = 12;
+    else if (nabi == "openvms")
+        newAbi = 13;
+    else
+        error("unrecognized OS ABI");
+
+    if (newAbi == abi) {
+        debug("current and requested OS ABIs are equal\n");
+        return;
+    }
+
+    hdr()->e_ident[EI_OSABI] = newAbi;
+    changed = true;
+}
+
 template<ElfFileParams>
 void ElfFile<ElfFileParamNames>::modifySoname(sonameMode op, const std::string & newSoname)
 {
@@ -1270,6 +1443,7 @@ void ElfFile<ElfFileParamNames>::modifyRPath(RPathOp op,
        string. */
     std::vector<std::string> neededLibs;
     auto dyn = (Elf_Dyn *)(fileContents->data() + rdi(shdrDynamic.sh_offset));
+    checkPointer(fileContents, dyn, sizeof(*dyn));
     Elf_Dyn *dynRPath = nullptr, *dynRunPath = nullptr;
     char * rpath = nullptr;
     for ( ; rdi(dyn->d_tag) != DT_NULL; dyn++) {
@@ -1345,8 +1519,8 @@ void ElfFile<ElfFileParamNames>::modifyRPath(RPathOp op,
     debug("new rpath is '%s'\n", newRPath.c_str());
 
 
-    if (rpath && newRPath.size() <= rpathSize) {
-        memcpy(rpath, newRPath.c_str(), newRPath.size() + 1);
+    if (newRPath.size() <= rpathSize) {
+        if (rpath) memcpy(rpath, newRPath.c_str(), newRPath.size() + 1);
         return;
     }
 
@@ -1652,6 +1826,38 @@ void ElfFile<ElfFileParamNames>::noDefaultLib()
     changed = true;
 }
 
+template<ElfFileParams>
+void ElfFile<ElfFileParamNames>::addDebugTag()
+{
+    auto shdrDynamic = findSectionHeader(".dynamic");
+
+    auto dyn = (Elf_Dyn *)(fileContents->data() + rdi(shdrDynamic.sh_offset));
+    for ( ; rdi(dyn->d_tag) != DT_NULL; dyn++) {
+        if (rdi(dyn->d_tag) == DT_DEBUG) {
+            return;
+        }
+    }
+    std::string & newDynamic = replaceSection(".dynamic",
+            rdi(shdrDynamic.sh_size) + sizeof(Elf_Dyn));
+
+    unsigned int idx = 0;
+    for ( ; rdi(((Elf_Dyn *) newDynamic.c_str())[idx].d_tag) != DT_NULL; idx++) ;
+    debug("DT_NULL index is %d\n", idx);
+
+    /* Shift all entries down by one. */
+    setSubstr(newDynamic, sizeof(Elf_Dyn),
+            std::string(newDynamic, 0, sizeof(Elf_Dyn) * (idx + 1)));
+
+    /* Add the DT_DEBUG entry at the top. */
+    Elf_Dyn newDyn;
+    wri(newDyn.d_tag, DT_DEBUG);
+    newDyn.d_un.d_val = 0;
+    setSubstr(newDynamic, 0, std::string((char *) &newDyn, sizeof(Elf_Dyn)));
+
+    this->rewriteSections();
+    changed = true;
+}
+
 template<ElfFileParams>
 void ElfFile<ElfFileParamNames>::clearSymbolVersions(const std::set<std::string> & syms)
 {
@@ -1682,6 +1888,9 @@ void ElfFile<ElfFileParamNames>::clearSymbolVersions(const std::set<std::string>
 }
 
 static bool printInterpreter = false;
+static bool printOsAbi = false;
+static bool setOsAbi = false;
+static std::string newOsAbi;
 static bool printSoname = false;
 static bool setSoname = false;
 static std::string newSoname;
@@ -1691,6 +1900,7 @@ static std::vector<std::string> allowedRpathPrefixes;
 static bool removeRPath = false;
 static bool setRPath = false;
 static bool addRPath = false;
+static bool addDebugTag = false;
 static bool printRPath = false;
 static std::string newRPath;
 static std::set<std::string> neededLibsToRemove;
@@ -1706,6 +1916,12 @@ static void patchElf2(ElfFile && elfFile, const FileContents & fileContents, con
     if (printInterpreter)
         printf("%s\n", elfFile.getInterpreter().c_str());
 
+    if (printOsAbi)
+        elfFile.modifyOsAbi(elfFile.printOsAbi, "");
+
+    if (setOsAbi)
+        elfFile.modifyOsAbi(elfFile.replaceOsAbi, newOsAbi);
+
     if (printSoname)
         elfFile.modifySoname(elfFile.printSoname, "");
 
@@ -1737,6 +1953,9 @@ static void patchElf2(ElfFile && elfFile, const FileContents & fileContents, con
     if (noDefaultLib)
         elfFile.noDefaultLib();
 
+    if (addDebugTag)
+        elfFile.addDebugTag();
+
     if (elfFile.isChanged()){
         writeFile(fileName, elfFile.fileContents);
     } else if (alwaysWrite) {
@@ -1778,6 +1997,8 @@ void showHelp(const std::string & progName)
   [--set-interpreter FILENAME]\n\
   [--page-size SIZE]\n\
   [--print-interpreter]\n\
+  [--print-os-abi]\t\tPrints 'EI_OSABI' field of ELF header\n\
+  [--set-os-abi ABI]\t\tSets 'EI_OSABI' field of ELF header to ABI.\n\
   [--print-soname]\t\tPrints 'DT_SONAME' entry of .dynamic section. Raises an error if DT_SONAME doesn't exist\n\
   [--set-soname SONAME]\t\tSets 'DT_SONAME' entry to SONAME.\n\
   [--set-rpath RPATH]\n\
@@ -1792,7 +2013,9 @@ void showHelp(const std::string & progName)
   [--replace-needed LIBRARY NEW_LIBRARY]\n\
   [--print-needed]\n\
   [--no-default-lib]\n\
+  [--no-sort]\t\tDo not sort program+section headers; useful for debugging patchelf.\n\
   [--clear-symbol-version SYMBOL]\n\
+  [--add-debug-tag]\n\
   [--output FILE]\n\
   [--debug]\n\
   [--version]\n\
@@ -1825,6 +2048,14 @@ int mainWrapped(int argc, char * * argv)
         else if (arg == "--print-interpreter") {
             printInterpreter = true;
         }
+        else if (arg == "--print-os-abi") {
+            printOsAbi = true;
+        }
+        else if (arg == "--set-os-abi") {
+            if (++i == argc) error("missing argument");
+            setOsAbi = true;
+            newOsAbi = resolveArgument(argv[i]);
+        }
         else if (arg == "--print-soname") {
             printSoname = true;
         }
@@ -1873,6 +2104,9 @@ int mainWrapped(int argc, char * * argv)
         else if (arg == "--print-needed") {
             printNeeded = true;
         }
+        else if (arg == "--no-sort") {
+            noSort = true;
+        }
         else if (arg == "--add-needed") {
             if (++i == argc) error("missing argument");
             neededLibsToAdd.insert(resolveArgument(argv[i]));
@@ -1901,6 +2135,9 @@ int mainWrapped(int argc, char * * argv)
         else if (arg == "--no-default-lib") {
             noDefaultLib = true;
         }
+        else if (arg == "--add-debug-tag") {
+            addDebugTag = true;
+        }
         else if (arg == "--help" || arg == "-h" ) {
             showHelp(argv[0]);
             return 0;
